from matplotlib import pyplot as plt


class ImageInfo:
    def __init__(self, image, label=None):
        self.image = image
        self.label = label

    def get_image_dimensions(self):
        print("self.image.shape", self.image.shape)
        height, width, channels = self.image.shape
        return height, width, channels

    def plot_image(self):
        plt.imshow(self.image)
        plt.axis('off')
        plt.title(f"label = {self.label}")
        plt.show()



import torch

from diamond.image_caption.utils import load_images_from_numpy
from diamond.user_classes import ImagesData, ImageDescriptionSpec

import os
from transformers import VisionEncoderDecoderModel, ViTImageProcessor, AutoTokenizer


class ImageDescription:
    def __init__(self,
                 images_data: ImagesData,
                 image_description_spec: ImageDescriptionSpec,
                 plot=True,
                 diamond_collector=None):
        self.images_data = images_data
        self.image_description_spec = image_description_spec
        self.plot = plot
        self.gen_kwargs = self.image_description_spec.gen_kwargs
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.label_dict = self.images_data.label_dict
        self.model, self.feature_extractor, self.tokenizer = None, None, None
        # TODO: later add support for labels (labels VS description?)
        self.diamond_collector = {} if diamond_collector is None else diamond_collector
        self.formatted_images = None
        self.format_images()
        self.load_model()

    def load_model(self):
        self.model, self.feature_extractor, self.tokenizer = get_image_captioning_models()

    def format_images(self):
        print("=>Formatting images...")
        self.formatted_images = load_images_from_numpy(self.images_data.raw_images)

    def predict_caption(self):
        new_images = []
        for image in self.formatted_images:
            if image.mode != "RGB":
                image = image.convert(mode="RGB")
            new_images.append(image)
        pixel_values = self.feature_extractor(images=new_images, return_tensors="pt").pixel_values
        pixel_values = pixel_values.to(self.device)
        output_ids = self.model.generate(pixel_values, **self.gen_kwargs)
        preds = self.tokenizer.batch_decode(output_ids, skip_special_tokens=True)
        preds = [pred.strip() for pred in preds]
        return preds

    def run(self):
        self.diamond_collector["image description"] = self.predict_caption()
        if self.plot:
            pass


def get_image_captioning_models():
    models_dir = "./models"

    # Check if models directory exists
    if not os.path.exists(models_dir):
        os.makedirs(models_dir)

    # Check if models exist in the models directory
    model_path = os.path.join(models_dir, "pytorch_model.bin")
    feature_extractor_path = os.path.join(models_dir, "processor.pt")
    tokenizer_path = os.path.join(models_dir, "tokenizer")

    if os.path.exists(model_path) and os.path.exists(feature_extractor_path) and os.path.exists(tokenizer_path):
        # Load the models if they exist
        model = VisionEncoderDecoderModel.from_pretrained(models_dir)
        feature_extractor = ViTImageProcessor.from_pretrained(models_dir)
        tokenizer = AutoTokenizer.from_pretrained(models_dir)
    else:
        # Download and save the models if they don't exist
        print("=> Downloading models...")
        model = VisionEncoderDecoderModel.from_pretrained("nlpconnect/vit-gpt2-image-captioning")
        feature_extractor = ViTImageProcessor.from_pretrained("nlpconnect/vit-gpt2-image-captioning")
        tokenizer = AutoTokenizer.from_pretrained("nlpconnect/vit-gpt2-image-captioning")

        model.save_pretrained(models_dir)
        feature_extractor.save_pretrained(models_dir)
        tokenizer.save_pretrained(models_dir)

    return model, feature_extractor, tokenizer



import random
from collections import Counter

import matplotlib
import pandas as pd
from chester.data_loader.webtext_data import load_data_pirates, load_data_king_arthur, load_data_chat_logs
from flatbuffers.builder import np
from sklearn.datasets import fetch_20newsgroups, fetch_openml

from chester.run.full_run import run_madcat
from chester.run.user_classes import Data, ModelRun, TimeSeriesHandler
import yfinance as yf
import ml_datasets

matplotlib.use('TkAgg')
target_column = 'target'


################################################################################################
df1 = load_data_pirates().assign(target='pirate').sample(300, replace=True)
df2 = load_data_king_arthur().assign(target='arthur').sample(300, replace=True)
df3 = load_data_chat_logs().assign(target='chat').sample(300, replace=True)
df = pd.concat([
    df1, df2,
    df3
])

# df['target'] = df['target'].apply(lambda x: 0 if "pirate" in x else 1)  # can do with or without
################################################################################################


################################################################################################
# url = "https://archive.ics.uci.edu/ml/machine-learning-databases/iris/iris.data"
# names = ['sepal-length', 'sepal-width', 'petal-length', 'petal-width', 'class']
# dataset = pd.read_csv(url, names=names)
# dataset.rename(columns={'class': 'target'}, inplace=True)
# df = dataset.sample(frac=1).reset_index(drop=True)
# df['target'] = df['target'].apply(lambda x: 0 if "Iris-setos" in x else 1)  # can do with or without
###############################################################################################

###############################################################################################
# Load the Boston Housing dataset. categorical
# boston = fetch_openml(name='boston', version=1)
# df = pd.DataFrame(boston.data, columns=boston.feature_names)
# df['target'] = boston.target
###############################################################################################

###############################################################################################
# df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/day.csv")
# df.rename(columns={'cnt': 'target'}, inplace=True)
###############################################################################################

###############################################################################################
# df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/camps.csv")
# df.rename(columns={'Competitors ': 'target'}, inplace=True)
###############################################################################################


###############################################################################################
# df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/daily_cinema.csv")
# df.rename(columns={'humidity': 'target'}, inplace=True)
##############################################################################################


###############################################################################################
def load_marvel_goodbad():
    heroes_information = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/heroes_information.csv")
    super_hero_powers = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/super_hero_powers.csv")
    df1 = pd.merge(heroes_information, super_hero_powers,
                   left_on='name', right_on='hero_names', how='left')

    df1.rename(columns={'Alignment': 'target'}, inplace=True)
    df1.drop(columns=['Unnamed: 0', 'name', 'hero_names'],
             inplace=True)

    # define a translation table to remove the specified characters

    def translate(str1):
        str1 = str1.lower()
        str1 = str1.replace('/', '')
        str1 = str1.replace('.', '')
        str1 = str1.replace('-', '')
        str1 = str1.replace('*', '')
        return str1

    # map the columns to their cleaned names using the translation table
    column_mapping = {col: translate(col) for col in df1.columns}
    df1 = df1.rename(columns=column_mapping)
    df1 = df1[df1['target'].isin(["good", "bad"])]

    return df1


# df = load_marvel_goodbad()
##############################################################################################

##############################################################################################
def load_maverl_dc():
    marvel = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/marvel-wikia-data.csv")
    dc = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/dc-wikia-data.csv")
    data1 = pd.concat([marvel, dc])
    data1.rename(columns={'ALIGN': 'target'}, inplace=True)
    data1.drop(columns=['urlslug', 'name'], inplace=True)
    return data1


# df = load_maverl_dc()
# print(df[0:10])
##############################################################################################

##############################################################################################
def load_autizm():
    data_aut = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/autism_prediction.csv")
    data_aut.rename(columns={'Class/ASD': 'target'}, inplace=True)
    return data_aut

# df = load_autizm()
##############################################################################################


###############################################################################################
# df = pd.read_csv("chester/model_training/models/chester_models/day.csv")
# df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/data/time_series_kaggle.csv")
# df = df[['date', 'store_nbr', 'family', 'sales']]

# top_stores = df.groupby('store_nbr').size().sort_values(ascending=False)[:5].index
# df = df[df['store_nbr'].isin(top_stores)]
# df.rename(columns={'sales': 'target'}, inplace=True)
###############################################################################################


################################################################################################
# from sklearn import datasets
# digits = datasets.load_digits()
# X = digits.images.reshape((len(digits.images), -1))
# df = pd.DataFrame(X)
# df.rename(columns={col: "feat_" + str(col) for col in df.columns}, inplace=True)
# df['target'] = digits.target
# df['target'] = "c_ " + df['target'].astype(str)
################################################################################################


################################################################################################
# categorical features
# import seaborn as sns
# df = sns.load_dataset("tips")
# df.rename(columns={'tip': target_column}, inplace=True)
################################################################################################

################################################################################################
# categorical
# import seaborn as sns
# df = sns.load_dataset("titanic")
# df.rename(columns={'survived': target_column}, inplace=True)
# df.drop(columns=['alive'], inplace=True)
###############################################################################################

###############################################################################################


################################################################################################
# df = pd.read_csv("chester/run_manual_chester/HR_Analytics.csv.csv")
# df.rename(columns={'Attrition': 'target'}, inplace=True)
###############################################################################################

###############################################################################################
# df = pd.read_csv("chester/run_manual_chester/flipkart_product.csv", encoding="ISO-8859-1")
# df = df[['Review', 'Summary', 'Rate']]
# df['Rate'] = "category: " + df['Rate'].astype(str)
# df.rename(columns={'Rate': 'target'}, inplace=True)
# df.dropna(subset=['target'], inplace=True)
# df = df.sample(5000)
###############################################################################################


###############################################################################################
# df1 = pd.read_csv("chester/run_manual_chester/Bank_Personal_Loan_Modelling.csv")
# df = df1.copy(deep=False)
# df.rename(columns={'Personal Loan': 'target'}, inplace=True)
#
# feature_types = {'numeric':
#                      ['ID', 'Age', 'Experience', 'Income', 'Family',
#                       'CCAvg', 'Education', 'Mortgage', 'Securities Account',
#                       'CD Account', 'Online', 'CreditCard'], 'boolean': [],
#                  'text': [], 'categorical': ['Education'], 'time': []}
###############################################################################################


###############################################################################################
# df1 = pd.read_csv("/Users/amitosi/PycharmProjects/chester/chester/run_manual_chester/Hotel Reservations.csv")
# df = df1.copy(deep=False)
# df.rename(columns={'booking_status': 'target'}, inplace=True)
# feature_types = {'numeric': ['no_of_adults', 'no_of_children', 'no_of_weekend_nights', 'no_of_week_nights',
#                              'required_car_parking_space', 'lead_time', 'arrival_year', 'arrival_month',
#                              'arrival_date', 'repeated_guest', 'no_of_previous_cancellations',
#                              'no_of_previous_bookings_not_canceled', 'avg_price_per_room',
#                              'no_of_special_requests'],
#                  'boolean': [],
#                  'text': [],
#                  'categorical': ['type_of_meal_plan', 'room_type_reserved', 'market_segment_type'], 'time': []}
###############################################################################################
# df1 = pd.read_csv("chester/run_manual_chester/loans_1.csv")
# df = df1.copy(deep=False)
# df.rename(columns={'Loan_Status': 'target'}, inplace=True)
# df.dropna(subset=['target'], inplace=True)
###############################################################################################

###############################################################################################


def generate_data(n_features, n_rows, target_type='binary'):
    if target_type == 'binary':
        # Create binary target column
        target = np.random.choice(['yes', 'no'], size=n_rows)
    elif target_type == 'multiclass':
        # Create multiclass target column
        target = np.random.choice(['class_1', 'class_2', 'class_3', 'class_4',
                                   'class_5', 'class_6', 'class_7', 'class_8',
                                   'class_9', 'class_10', 'class_11', 'class_12'], size=n_rows)
    else:
        raise ValueError("Invalid target_type. Must be either 'binary' or 'multiclass'.")

    # Create feature categorical columns
    features = {}
    for i in range(n_features):
        feature = np.random.choice(['A', 'B', 'C', 'D'], size=n_rows)
        features[f'feature_{i}'] = feature

    # Create pandas DataFrame
    df = pd.DataFrame(features)
    df['target'] = target

    return df


# df = generate_data(1, 100, target_type='binary')
# df = generate_data(30, 100, target_type='multiclass')
###############################################################################################


def load_ex2():
    newsgroups_train = fetch_20newsgroups(subset='train')
    df = pd.DataFrame(newsgroups_train.data, columns=['text'])
    y = newsgroups_train.target
    df['target'] = y
    category_counts = Counter(y)
    top_3_categories = category_counts.most_common(4)
    top_3_categories = [cat for cat, count in top_3_categories]
    df = df[df.target.isin(top_3_categories)].sample(1500)
    df['target'] = "c_ " + df['target'].astype(str)
    return df


def load_ex3():
    train_data, _ = ml_datasets.dbpedia()
    train_data = pd.DataFrame(train_data, columns=["text", "target"])
    train_data['target'] = "c_: " + train_data['target'].astype(str)
    return train_data


def load_ex4():
    train_data, _ = ml_datasets.cmu()
    train_data = pd.DataFrame(train_data, columns=["text", "target"])
    train_data['target'] = "category: " + train_data['target'].astype(str)
    train_data = train_data[train_data['target'].isin(train_data['target'].value_counts()[:5].index)]
    return train_data


def load_ex5():
    train_data, _ = ml_datasets.quora_questions()
    train_data = [(a, b, c) for ((a, b), c) in train_data]
    train_data = pd.DataFrame(train_data, columns=["text1", "text2", "target"])
    return train_data


# load data
# df = load_ex2()
# df = load_ex3().sample(1000)
# df = load_ex4().sample(1000)
# df = load_ex5().sample(900)


def load_yaho(tickers=None, start_date='2010-01-01', end_date='2023-02-15'):
    # Download stock data for all tickers and concatenate them into a single dataframe
    if tickers is None:
        tickers = ['AAPL', 'MSFT', 'GOOG']
    df_list = []
    for ticker in tickers:
        ticker_df = yf.download(ticker, start=start_date, end=end_date)
        ticker_df["id"] = ticker
        df_list.append(ticker_df[['Close', 'id']])

    # Combine all dataframes into one and reset the index
    df = pd.concat(df_list).reset_index()

    # Rename the columns to match the desired output
    df = df.rename(columns={'Date': 'date', 'Close': 'value'})

    # Rearrange the columns to match the desired output
    df = df[['id', 'date', 'value']]
    df.rename(columns={'value': 'target'}, inplace=True)

    return df


# df = load_yaho()
# print("df shape", df.shape)
# print("df cols", df.columns)
#
madcat_collector = run_madcat(Data(df=df, target_column='target'),
                              is_feature_stats=True,
                              # time_series_handler=TimeSeriesHandler(id_cols=["id"]),
                              is_pre_model=True,
                              is_model_training=True,
                              model_run=ModelRun(n_models=2),
                              is_post_model=True, is_model_weaknesses=True,
                              plot=True,
                              # feature_types={'numeric': [], 'boolean': [], 'text': ['Campaign Name'],
                              #                'categorical': ['Campaign Name'], 'time': ['Date'], 'id': []}
                              )

# output_collector = run_madcat(
#     Data(df=df, target_column='target'),
#     model_run=ModelRun(n_models=10),
#     is_feature_stats=True,
#     is_pre_model=True,
#     # feature_types=feature_types,
#     is_model_training=True, is_post_model=True, is_model_weaknesses=True
# )



import tamtam
import pandas as pd

from tamtam.ab_info.ab_class import ABInfo
from tamtam.run import run
from tamtam.user_class.user_class import ABData, TestInfo

# read
# df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/tamtam/data/cookie_cats.csv")
# df["side"] = df["version"].apply(lambda x: "B" if x == "gate_30" else "A")
# df.drop(columns=["version"], inplace=True)
# ab_data = ABData(df)
# test_info = TestInfo(side_col="side", metrics=['retention_1', 'retention_7'], id_cols=['userid'],
#                      feature_cols=['sum_gamerounds'])

# df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/tamtam/data/ab_data.csv")
# countries = pd.read_csv("/Users/amitosi/PycharmProjects/chester/tamtam/data/countries.csv")
# df = pd.merge(df, countries, on="user_id", how="left")
# df["side"] = df["group"].apply(lambda x: "B" if x == "treatment" else "A")
# df.drop(columns=["group"], inplace=True)
#
# # Prepare
# ab_data = ABData(df.sample(5000))
# test_info = TestInfo(side_col="side", metrics=['converted'], id_cols=['user_id'],
#                      feature_cols=['landing_page', 'country'],
#                      date_col="timestamp")
#
# # run
# run(ab_data=ab_data, test_info=test_info)

# df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/tamtam/data/AB_Test_Results.csv")
# df["side"] = df["VARIANT_NAME"].apply(lambda x: "B" if x == "variant" else "A")
# df.drop(columns=["VARIANT_NAME"], inplace=True)
#
# # Prepare
# ab_data = ABData(df.sample(5000))
# test_info = TestInfo(side_col="side", metrics=['REVENUE'], id_cols=['USER_ID'])

#
# df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/tamtam/data/abtest_kaggle_3.csv")
# df["side"] = df["group"].apply(lambda x: "A" if x == "control" else "B")
# df.drop(columns=["group"], inplace=True)
# df["ctr"] = df["clicks"] / df["views"]
# df.drop(columns=["clicks"], inplace=True)
#
# # Prepare
# ab_data = ABData(df.sample(5000))
# test_info = TestInfo(side_col="side", metrics=['ctr'], id_cols=['user_id'], weight_col='views')
#
# # run
# run(ab_data=ab_data, test_info=test_info)

df = pd.read_csv("/Users/amitosi/PycharmProjects/chester/tamtam/data/lalonde.csv")
# read data
df["side"] = df["treat"].apply(lambda x: "B" if x == 1 else "A")
df.drop(columns=["treat"], inplace=True)

# test specifications
ab_data = ABData(df)
test_info = TestInfo(
    side_col="side", metrics=['re74', 're75', 're78'],
    feature_cols=['age', 'educ', 'black', 'hispan', 'married', 'nodegree'])

run(ab_data=ab_data, test_info=test_info)



from typing import List

import numpy as np
from prettytable import PrettyTable
from sklearn.exceptions import UndefinedMetricWarning

from chester.feature_stats.utils import create_pretty_table
from chester.model_training.data_preparation import CVData
from chester.model_training.data_preparation import Parameter
from chester.model_training.models.chester_models.base_model_utils import is_metric_higher_is_better
from chester.model_training.models.chester_models.catboost.catboost_model import CatboostModel
from chester.model_training.models.chester_models.hp_generator import HPGenerator
from chester.zero_break.problem_specification import DataInfo


def train_catboost(X_train, y_train, parameters: list, data_info: DataInfo):
    """
    Trains a baseline model using the given parameters.
    :param X_train: The training data features (unused in this function)
    :param y_train: The training data labels
    :param parameters: parameters list
    :return: A trained baseline model
    """
    model = CatboostModel(parameters, data_info)
    model.fit(X_train, y_train)
    return model


def predict_catboost(model, X):
    """
    Makes predictions using a baseline model
    :param model: trained baseline model
    :param X: The data to make predictions on (unused in this function)
    :return: A list of predictions
    """
    return model.transform(X)


def catboost_with_outputs(cv_data: CVData,
                          metrics: list,
                          target_col: str,
                          parameters: list,
                          data_info: DataInfo,
                          ):
    results = []
    problem_type = data_info.problem_type_val
    for i, (X_train, y_train, X_test, y_test) in enumerate(cv_data.format_splits()):
        model = train_catboost(X_train, y_train, parameters=parameters, data_info=data_info)
        prediction = predict_catboost(model, X_test)
        prediction_train = predict_catboost(model, X_train)
        scores = calculate_catboost_metrics_scores(y_test, prediction, metrics, problem_type)
        results.append({'type': 'test', 'fold': i, **scores})
        scores = calculate_catboost_metrics_scores(y_train, prediction_train, metrics, problem_type)
        results.append({'type': 'train', 'fold': i, **scores})
    return results, model


class_name_dict = {
    'LinearRegression': 'Linear Model',
    'LGBMRegressor': 'Light GBM',
    'LogisticRegression': 'Logistic Regression',
    'Sequential': 'Sequential Model',
    'RandomForestRegressor': 'Random Forest',
    'XGBRegressor': 'XGBoost',
    'SVR': 'Support Vector Regression',
    'KNeighborsRegressor': 'K-Neighbors Regression',
    'DecisionTreeRegressor': 'Decision Tree',
    'MLPRegressor': 'Multi-layer Perceptron',
    'AdaBoostRegressor': 'AdaBoost',
    'GradientBoostingRegressor': 'Gradient Boosting',
    'ExtraTreesRegressor': 'Extra Trees',
    'BaggingRegressor': 'Bagging',
    'CatBoostRegressor': 'CatBoost',
    'Lasso': 'Lasso',
    'Ridge': 'Ridge',
    'ElasticNet': 'Elastic Net',
    'PassiveAggressiveRegressor': 'Passive Aggressive'
}


def get_model_name(model):
    model_name = class_name_dict.get(model.__class__.__name__, model.__class__.__name__)
    return model_name


def compare_models(results):
    all_results = [(pd.DataFrame(result), model) for result, model in results]
    metric_name = [col for col in all_results[0][0].columns if col not in ['type', 'fold']][0]
    sort_ascending = is_metric_higher_is_better(metric_name)
    best_result = None
    best_model = None
    best_value = None
    for (result, model) in all_results:
        test_result = result[result['type'] == 'test'].groupby('fold').mean(numeric_only=True).reset_index()
        mean_value = test_result[metric_name].mean()
        if best_value is None or \
                (sort_ascending and mean_value > best_value) \
                or (not sort_ascending and mean_value < best_value):
            best_value = mean_value
            best_result = result
            best_model = model
    return best_result, best_model


def calculate_average(df):
    df_grouped = df.drop(columns=['fold']).groupby(['type', 'model'], as_index=False).mean(numeric_only=True)
    return df_grouped


import pandas as pd
import matplotlib.pyplot as plt


def visualize_performance(df, with_baseline=True):
    addition = 'excluding baseline model' if not with_baseline else ''

    # Create pivot table to summarize mean performance metric by type and model
    metric_columns = [col for col in df.columns if col not in ["type", "model"]]
    pivot = df.pivot(index="model", columns="type", values=metric_columns)

    if not with_baseline:
        pivot = df[df['model'] != 'BaselineModel'].pivot(index="model", columns="type", values=metric_columns)

    # Plot bar chart to compare mean performance metric by model
    # fig, axes = plt.subplots(nrows=len(metric_columns), ncols=1, figsize=(12, 10), sharex=True)
    fig, axes = plt.subplots(nrows=len(metric_columns), ncols=1, figsize=(12, 10), sharex=True)

    for i, metric in enumerate(metric_columns):
        pivot[metric].plot.bar(ax=axes[i], rot=0)
        axes[i].set_ylabel(None)
        axes[i].set_title("Comparison of {} by Model {}".format(metric, addition))

    plt.xlabel(None)
    plt.ylabel(None)
    plt.tight_layout()
    plt.show()
    plt.close()

    return pivot


def compare_best_models(results, plot_results=True):
    from chester.util import ReportCollector, REPORT_PATH
    rc = ReportCollector(REPORT_PATH)

    all_results = []
    all_results_with_models = []
    for res in results:
        result, model = res
        result_organized = pd.DataFrame(result)
        result_organized['model'] = get_model_name(model)
        all_results.append(result_organized)
        all_results_with_models.append((result_organized, model))

    metrics_results = calculate_average(pd.concat(all_results))
    if plot_results:
        print("Model results - comparing the best out of each type")
        visualize_performance(metrics_results, with_baseline=True)
        print("Model results - comparing the best out of each type, excluding baseline model")
        visualize_performance(metrics_results, with_baseline=False)

    rc.save_object(create_pretty_table(metrics_results),
                   text="Model results compared - showing the best out of each type after CV & HP tuning: ")

    metric_name = [col for col in all_results[0].columns if col not in ['type', 'fold', 'model']][0]
    sort_ascending = is_metric_higher_is_better(metric_name)
    best_result = None
    best_model = None
    best_value = None
    for (result, model) in all_results_with_models:
        test_result = result[result['type'] == 'test'].groupby('fold').mean(numeric_only=True).reset_index()
        mean_value = test_result[metric_name].mean()
        if best_value is None or \
                (sort_ascending and mean_value > best_value) \
                or (not sort_ascending and mean_value < best_value):
            best_value = mean_value
            best_result = result
            best_model = model
    print(f"Optimized {metric_name}, with best value: {round(best_value, 2)}. ")
    return best_result, best_model


default_parameters = {
    'iterations': 1000,
    'learning_rate': 0.03,
    'depth': 6,
    'random_seed': 42,
    'verbose': False
}

# For regression
regression_parameters = default_parameters.copy()
regression_parameters.update({'loss_function': 'RMSE', 'eval_metric': 'RMSE'})

# For binary classification
binary_classification_parameters = default_parameters.copy()
binary_classification_parameters.update({'loss_function': 'Logloss', 'eval_metric': 'AUC'})

# For multiclass classification
multiclass_classification_parameters = default_parameters.copy()
multiclass_classification_parameters.update({'loss_function': 'MultiClass', 'eval_metric': 'Accuracy'})


def generate_catboost_configs(k: int = 10) -> List[List[Parameter]]:
    hp_generator = HPGeneratoCatboost(n_models=k)
    parameter_format = hp_generator.hp_format(hp_generator.generate_configs())
    return parameter_format


class HPGeneratoCatboost(HPGenerator):
    def __init__(self, best_practice_configs: list = None,
                 categorical_sample_configs: list = None,
                 n_models=9,
                 best_practice_prob=0):
        super().__init__(best_practice_configs, categorical_sample_configs, n_models, best_practice_prob)
        self.best_practice_configs = []

    def generate_random_config(self) -> dict:
        config = {}
        # Sampling depth
        depth = np.random.randint(2, 6)
        config['depth'] = depth
        # Sampling learning rate
        learning_rate = np.random.uniform(0.01, 0.2)
        config['learning_rate'] = learning_rate
        # Sampling iterations
        iterations = np.random.randint(100, 1000)
        config['iterations'] = iterations
        # Sampling l2 regularization
        l2_reg = 10 ** np.random.uniform(-5, 5)
        config['l2_leaf_reg'] = l2_reg
        # Sampling random strength
        random_strength = np.random.uniform(1, 100)
        config['random_strength'] = random_strength
        # Sampling bagging temperature
        bagging_temp = np.random.uniform(0, 1)
        config['bagging_temperature'] = bagging_temp
        config['verbose'] = False
        return config


def calculate_catboost_metric_score(y_true, y_pred, metric, problem_type_val):
    import warnings
    warnings.filterwarnings("ignore", category=UndefinedMetricWarning)
    metric_name = metric.__name__
    if problem_type_val in ["Multiclass classification"]:
        y_pred = [item[0] for item in y_pred]
    try:
        return metric_name, metric(y_true, y_pred)
    except:
        try:
            if problem_type_val in ["Binary regression", "Binary classification"]:
                return metric_name, metric(y_true, y_pred, average='binary')
            elif problem_type_val == "Multiclass classification":
                return metric_name, metric(y_true, y_pred, average='macro')
        except:
            pass
    return None, None


def calculate_catboost_metrics_scores(y, prediction, metrics_list, problem_type=None):
    results = {}
    for metric in metrics_list:
        metric_name, metric_value = calculate_catboost_metric_score(y, prediction, metric, problem_type)
        results[metric_name] = metric_value
    return results



import numpy as np
import pandas as pd

from tamtam.user_class.user_class import ABData, TestInfo
from tamtam.utils.column_utils import Columns


class ABInfo:
    def __init__(self, ab_data: ABData, test_info: TestInfo):
        self.ab_data = ab_data
        self.test_info = test_info
        self.df = self.ab_data.df
        self.run()  # validate cols
        # additional columns to calculate
        self.control_name = "A"
        self.treatment_name = "B"
        self.trimmed, self.trimmed_cols, self.trimmed_col_by_metric = self.trim()

    def add_dummies(self):
        if not self.test_info.is_weight_col():
            self.df[Columns.weight] = 1
        if not self.test_info.is_id_cols():
            self.df[Columns.id] = self.df.index

    def aggregate(self):
        self.add_dummies()
        weight_col = self.test_info.get_weight_col() or Columns.weight
        id_cols = self.test_info.get_id_cols() or Columns.id
        df = self.df.copy(deep=False)
        side_col = self.test_info.side_col
        metrics = self.test_info.metrics
        df[metrics] = df[metrics].multiply(df[side_col].map({'A': -1, 'B': 1}), axis=0)

        def weighted_avg(group):
            weights = group[weight_col]
            weighted_metrics = group[metrics].multiply(weights, axis=0)
            avg = weighted_metrics.sum() / weights.sum()
            weight_sum = weights.sum()
            return pd.Series({**avg, weight_col: weight_sum})

        result = df.groupby(id_cols).apply(weighted_avg)
        result = result.reset_index()

        return result

    @staticmethod
    def weighted_percentile(values, weights, percentile):
        """Compute the weighted percentile of a list of values."""
        sorter = np.argsort(values)
        values = values[sorter]
        weights = weights[sorter]
        weighted_quantiles = np.cumsum(weights) - 0.5 * weights
        percentile_values = np.interp(percentile, weighted_quantiles / np.sum(weights), values)
        return percentile_values

    def trim(self):
        aggregated = self.aggregate()

        trimming_list = self.test_info.trimming_list  # [0, 0.01, .... ]
        weight_col = self.test_info.get_weight_col() or Columns.weight  # string
        metrics = self.get_metric_cols()

        trimmed_cols = []
        trimmed_col_by_metric = {}
        for m in metrics:
            trimmed_col_by_metric[m] = []
            for p in trimming_list:
                trimmed_col = f"{m}_trim_{100 * p:.2f}"
                q = [p / 2, 1 - p / 2]
                trimmed_vals = self.weighted_percentile(aggregated[m], aggregated[weight_col], q)
                trimmed = np.clip(aggregated[m], *trimmed_vals)

                if len(np.unique(trimmed)) == 1:
                    pass
                else:
                    aggregated[trimmed_col] = trimmed
                    trimmed_cols.append(trimmed_col)
                    trimmed_col_by_metric[m].extend([trimmed_col])

        return aggregated, trimmed_cols, trimmed_col_by_metric

    def get_weight_col(self):
        return self.test_info.get_weight_col()

    def validate_cols(self):
        all_cols = self.test_info.get_all_cols()
        df_cols = list(self.df.columns)
        # Check if all columns in all_cols exist in df_cols
        diff_cols = set(all_cols) - set(df_cols)
        if diff_cols:
            # Raise an error with the list of columns not in df
            error_msg = f"The following columns are not in the provided dataframe: {list(diff_cols)}"
            raise ValueError(error_msg)
        else:
            # All columns exist in df, so pass
            pass

    def cast_cols(self):
        metrics = self.get_metric_cols()
        df = self.df
        df[metrics] = df[metrics].astype(float)
        self.df = df

    def get_lead_metric(self):
        return self.test_info.metrics[0]

    def get_metric_cols(self):
        return self.test_info.metrics

    def run(self):
        self.validate_cols()
        self.cast_cols()



from chester.features_engineering.time_series.ts_feature_extraction import TimeSeriesFeatureExtraction
from chester.run.user_classes import TimeSeriesHandler
from chester.zero_break.problem_specification import DataInfo


class TimeSeriesFeaturesExtraction:
    def __init__(self,
                 time_series_handler: TimeSeriesHandler = None,
                 data_info: DataInfo = None
                 ):
        self.time_series_handler = time_series_handler or TimeSeriesHandler()
        self.data_info = data_info
        self.cols = self.data_info.feature_types_val["time"]

    def run(self):
        for col in self.cols:
            ts_fe = TimeSeriesFeatureExtraction(
                data_info=self.data_info,
                time_series_handler=self.time_series_handler,
                col_name=col,
                column=self.data_info.data[col]
            )
            ts_fe.run()
            self.data_info = ts_fe.data_info
            # print("updated data info after ts", self.data_info)



import random
import math
import matplotlib.pyplot as plt
import numpy as np
import torch
from torch.utils.data import TensorDataset, DataLoader
from tqdm import tqdm
import pandas as pd


class ImagesData:
    def __init__(self, images, labels, validation_prop, image_shape, label_dict):
        self.images = images
        self.raw_images = self.images
        self.labels = labels
        self.validation_prop = validation_prop
        self.image_shape = image_shape
        self.label_dict = label_dict
        self.problem_type = self.get_problem_type()
        # reshape
        # Convert to ndarray and reshape
        if isinstance(self.images, pd.DataFrame):
            self.images = self.images.to_numpy()
        self.label_hanlder()

        self.is_colored = True if len(self.image_shape) > 2 else False
        if self.is_colored:
            self.images_to_show = np.transpose(self.images, (0, 2, 3, 1))
        else:
            self.images_to_show = np.transpose(self.images, (0, 1, 2))
        self.images_train, self.labels_train, self.images_val, self.labels_val = self.split()
        if self.is_colored:
            print("Colored image")
        else:
            print("Grayscale image")

    def label_hanlder(self):
        if isinstance(self.labels, pd.DataFrame):
            self.labels = self.labels.to_numpy()

    def split(self):
        assert 0 <= self.validation_prop < 0.8, "validation proportion should be in range (0, 0.8)"
        try:
            self.images = self.images.reshape((-1,) + self.image_shape)
            num_val = int(self.validation_prop * len(self.images))
            indices = np.random.permutation(len(self.images))
            val_indices, train_indices = indices[:num_val], indices[num_val:]
            images_train, labels_train = self.images[train_indices], self.labels[train_indices]
            images_val, labels_val = self.images[val_indices], self.labels[val_indices]
            return images_train, labels_train, images_val, labels_val
        except:
            num_val = int(self.validation_prop * len(self.images))
            indices = np.random.permutation(len(self.images))
            val_indices, train_indices = indices[:num_val], indices[num_val:]
            images_train, labels_train = self.images[train_indices], self.labels[train_indices]
            images_val, labels_val = self.images[val_indices], self.labels[val_indices]
            return images_train, labels_train, images_val, labels_val

    def create_data_loaders(self, batch_size):
        train_dataset = TensorDataset(torch.from_numpy(self.images_train), torch.from_numpy(self.labels_train))
        train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True)

        val_dataset = TensorDataset(torch.from_numpy(self.images_val), torch.from_numpy(self.labels_val))
        val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False)

        # add tqdm progress bar to train_loader
        train_loader = tqdm(train_loader, total=len(train_loader))

        return train_loader, val_loader

    def get_splits(self):
        return self.images_train, self.labels_train, self.images_val, self.labels_val

    @staticmethod
    def get_problem_type():
        return "classification"

    def plot_images(self):
        num_images = len(self.images_to_show)
        if num_images > 100:
            image_indices = range(num_images - 100, num_images)
        else:
            image_indices = random.sample(range(num_images), num_images)

        num_cols = int(math.floor(math.sqrt(len(image_indices))))
        num_rows = int(math.ceil(len(image_indices) / num_cols))

        fig, ax = plt.subplots(num_rows, num_cols, figsize=(12, 12))
        for i, index in enumerate(image_indices):
            row = i // num_cols
            col = i % num_cols
            ax[row, col].imshow(self.images_to_show[index])
            ax[row, col].axis('off')
            if self.labels is not None:
                ax[row, col].set_title(str(self.labels[index]))
        plt.show()


class ImagesAugmentationInfo:

    def __init__(self, aug_types=None, aug_prop=0.05):
        self.aug_types = aug_types
        self.aug_prop = aug_prop
        # calc
        if self.aug_types is None:
            self.aug_types = ["zoom", "rotate"]


class ImageModel:
    def __init__(self,
                 network_name="EfficientNetB0",
                 remove_num_last_layers=1,
                 batch_size=64,
                 num_epochs=3,
                 optimizer_params=None,
                 dropout=0.5
                 ):
        self.network_name = network_name
        self.remove_last_layers_num = remove_num_last_layers
        self.batch_size = batch_size
        self.num_epochs = num_epochs
        self.optimizer_params = optimizer_params
        self.dropout = dropout
        if self.optimizer_params is None:
            self.optimizer_params = {"lr": 0.001, 'weight_decay': 0.0001}
        self.network_parameters = self.optimizer_params.copy()
        self.network_parameters["_model_name"] = self.network_name
        self.network_parameters["_num_epochs"] = self.num_epochs
        self.network_parameters["_batch_size"] = self.batch_size
        self.network_parameters["_dropout"] = self.dropout
        self.network_parameters["_remove_last_layers_num layers"] = self.remove_last_layers_num


class ImageModels:
    def __init__(self, image_model_list=None):
        self.image_model_list = image_model_list
        if image_model_list is None:
            self.image_model_list = [ImageModel()]
        self.n_models = len(self.image_model_list)


class ImagePostModelSpec:
    def __init__(self, plot,
                 is_compare_models=True,
                 is_confusion_matrix=True,
                 is_precision_recall=True):
        self.plot = plot
        self.is_compare_models = is_compare_models
        self.is_confusion_matrix = is_confusion_matrix
        self.is_precision_recall = is_precision_recall


class ImageDescriptionSpec:
    def __init__(self, max_length=16, num_beams=4):
        self.max_length = max_length
        self.num_beams = num_beams
        self.gen_kwargs = {"max_length": self.max_length, "num_beams": self.num_beams}



import pandas as pd

from chester.features_engineering.time_series.cyclic_features_utils import CyclicFeatures
from chester.features_engineering.time_series.feature_elimination_utils import FeatureEliminationUtils
from chester.features_engineering.time_series.frequencies_utils import FrequenciesFeatures
from chester.features_engineering.time_series.get_time_freqeuency_utils import TimeFrequencyDecider
from chester.features_engineering.time_series.moving_metric_utils import MovingMetric
from chester.features_engineering.time_series.static_features_utils import TSStaticFeatures
from chester.run.user_classes import TimeSeriesHandler
from chester.zero_break.problem_specification import DataInfo


class TimeSeriesFeatureExtraction:
    def __init__(self,
                 column,
                 col_name,
                 time_series_handler: TimeSeriesHandler = None,
                 data_info: DataInfo = None
                 ):
        self.col_name = col_name
        self.column = column
        self.time_series_handler = time_series_handler or TimeSeriesHandler()
        self.data_info = data_info
        # save
        self.time_frequency = self.time_series_handler.time_frequency or self._get_time_frequency()
        self.time_series_handler.time_frequency = self.time_frequency

    def static_features(self):
        stat_features = TSStaticFeatures(column=self.column,
                                         col_name=self.col_name,
                                         time_series_handler=self.time_series_handler,
                                         data_info=self.data_info)
        sf_df, names = stat_features.run()
        sf_df = FeatureEliminationUtils(df=sf_df[names]).run()  # Elimination for the relevant features only
        self.data_info.data = pd.concat([self.data_info.data, sf_df], axis=1)  # update data
        self.data_info.feature_types_val["categorical"] = list(
            self.data_info.feature_types_val["categorical"])  # convert set to list
        self.data_info.feature_types_val["categorical"].extend(list(sf_df.columns))  # update features

    # return (interval, counts)
    def _get_time_frequency(self):
        decider = TimeFrequencyDecider(column=self.column,
                                       col_name=self.col_name,
                                       time_series_handler=self.time_series_handler,
                                       data_info=self.data_info)
        return decider.run()

    def target_lags(self):
        mm = MovingMetric(column=self.column,
                          col_name=self.col_name,
                          time_series_handler=self.time_series_handler,
                          data_info=self.data_info)
        target_df, names = mm.run()
        target_df = FeatureEliminationUtils(df=target_df[names]).run()  # Elimination for the relevant features only
        self.data_info.data = pd.concat([self.data_info.data, target_df], axis=1)  # update data
        self.data_info.feature_types_val["numeric"] = list(
            self.data_info.feature_types_val["numeric"])  # convert set to list
        self.data_info.feature_types_val["numeric"].extend(target_df.columns)  # update features

    def freq_features(self):
        ff = FrequenciesFeatures(column=self.column,
                                 col_name=self.col_name,
                                 time_series_handler=self.time_series_handler,
                                 data_info=self.data_info)
        df_ff, names = ff.run()
        df_ff = FeatureEliminationUtils(df=df_ff[names]).run()  # Elimination for the relevant features only
        self.data_info.data = pd.concat([self.data_info.data, df_ff], axis=1)  # update data
        self.data_info.feature_types_val["numeric"] = list(
            self.data_info.feature_types_val["numeric"])  # convert set to list
        self.data_info.feature_types_val["numeric"].extend(df_ff.columns)  # update features

    def cyclic_features(self):
        cf = CyclicFeatures(column=self.column,
                            col_name=self.col_name,
                            time_series_handler=self.time_series_handler,
                            data_info=self.data_info)
        df_cf, names = cf.run()
        df_cf = FeatureEliminationUtils(df=df_cf[names]).run()  # Elimination for the relevant features only
        self.data_info.data = pd.concat([self.data_info.data, df_cf], axis=1)
        self.data_info.feature_types_val["numeric"] = list(
            self.data_info.feature_types_val["numeric"])  # convert set to list
        self.data_info.feature_types_val["numeric"].extend(df_cf.columns)  # update features

    def run(self):
        self.static_features()  # return df, feature names
        self.cyclic_features()
        if 'regression' in self.data_info.problem_type_val.lower():
            self.target_lags()
        self.freq_features()



import numpy as np
import pandas as pd

from chester.run.user_classes import TimeSeriesHandler
from chester.zero_break.problem_specification import DataInfo


class CyclicFeatures:
    def __init__(self, column,
                 col_name,
                 time_series_handler: TimeSeriesHandler = None,
                 data_info: DataInfo = None):
        self.date_col_name = col_name
        self.column = column
        self.time_series_handler = time_series_handler
        self.data_info = data_info
        self.df = self.data_info.data

    def extract_relevant_date_cols(self):
        optional_static_features = ['ts_minutes', 'ts_hours', 'ts_dayinmonth', 'ts_month',
                                    'ts_quarter', 'ts_year']
        all_features = self.df.columns
        relevant_features = [feat for feat in all_features if feat.startswith(tuple(optional_static_features))]
        date_cols = [col for col in relevant_features if self.date_col_name in col]
        return date_cols

    def calculate_cyclic_features(self):
        cols = self.extract_relevant_date_cols()
        out_cols = []
        for col in cols:
            # Calculate the sine and cosine of the date column
            sin_col_name = f"{col}_sin"
            cos_col_name = f"{col}_cos"
            sin_values = self.df[col].apply(lambda x: np.sin(2 * np.pi * x / self.df[col].max()))
            cos_values = self.df[col].apply(lambda x: np.cos(2 * np.pi * x / self.df[col].max()))

            # Add the new columns to the DataFrame
            self.df[sin_col_name] = sin_values
            self.df[cos_col_name] = cos_values
            out_cols.append(sin_col_name)
            out_cols.append(cos_col_name)
        return out_cols

    def run(self):
        names = self.calculate_cyclic_features()
        return self.df, names



from chester.run.full_run import run_madcat
from chester.run.user_classes import Data, ModelRun
from tamtam.ab_feature_analysis.ab_catboost import ABCatboostPlotter
from tamtam.ab_feature_analysis.ab_partial_plot import ABPartialPlot
from tamtam.ab_feature_analysis.ab_tree_class import ABTreePlot
from tamtam.ab_info.ab_class import ABInfo
from tamtam.user_class.user_class import TestInfo


class ABFeatureAnalysis:
    def __init__(self, ab_info: ABInfo, test_info: TestInfo):
        self.ab_info = ab_info
        self.test_info = test_info
        self.df = self.ab_info.df
        self.feature_cols = self.test_info.feature_cols or []
        # df[metrics] = df[metrics].multiply(df[side_col].map({'A': -1, 'B': 1}), axis=0)

    def plot_tree(self, metric):
        ab_tree = ABTreePlot(ab_info=self.ab_info, test_info=self.test_info, metric=metric)
        ab_tree.run()

    def plot_catboost(self, metric):
        ab_catboost = ABCatboostPlotter(ab_info=self.ab_info, test_info=self.test_info, metric=metric)
        ab_catboost.run()

    def partial_plot(self, metric):
        ab_partial_plot = ABPartialPlot(ab_info=self.ab_info, test_info=self.test_info, metric=metric)
        ab_partial_plot.run()

    def chester_run(self, metric):
        # prepare data
        df = self.df.copy()
        df[metric] = df[metric].multiply(df[self.test_info.side_col].map({'A': -1, 'B': 1}), axis=0)

        df = df[[metric] + self.feature_cols]
        df.rename(columns={metric: 'target'}, inplace=True)
        # run chester
        run_madcat(data_spec=Data(df=df, target_column='target'), model_run=ModelRun(n_models=3),
                   is_feature_stats=False, is_pre_model=True, is_model_weaknesses=False)

    def run_single(self, metric):
        # self.plot_tree(metric=metric)
        # self.plot_catboost(metric=metric)
        # self.partial_plot(metric=metric)
        self.chester_run(metric=metric)

    def run(self):
        metrics = self.ab_info.get_metric_cols()
        for metric in metrics:
            print(f"==========================> Training a model to predict diff in {metric}")
            self.run_single(metric=metric)



# goal: input to image
from PIL import Image
import numpy as np
#
# from diamond.image_caption.utils import *
#
# paths = ["/Users/amitosi/PycharmProjects/chester/chester/data/weather/sunrise336.jpg",
#          "/Users/amitosi/PycharmProjects/chester/chester/data/weather/rain118.jpg"]
# images = load_images_from_path(paths)
# numpy_images = [load_from_numpy(image) for image in images]
# # img_np = np.array(image)
# images = load_images_from_numpy(numpy_images)
# preidictions = predict_caption(images)
# print(preidictions)



import warnings

from chester.run.feature_attention_utils import FeatureTypeCategorizer
from chester.zero_break.problem_specification import DataInfo

warnings.filterwarnings("ignore", category=UserWarning, module="lightgbm")
import logging

logging.basicConfig(level=logging.WARNING)
from chester.cleaning import cleaning_func as cln
from chester.preprocessing import preprocessing_func as pp

import pandas as pd


class Data:
    def __init__(self, df: pd.DataFrame, target_column: str = None):
        self.df = df
        self.target_column = target_column


class FeatureTypes:
    def __init__(self, feature_types: dict):
        self.feature_types = feature_types

    def fix_feature_types(self):
        self.feature_types = FeatureTypeCategorizer(self.feature_types).categorize_features()

    def update_data_info(self, data_info: DataInfo):
        self.fix_feature_types()
        data_info.feature_types_val = self.feature_types
        return data_info


class TextHandler:
    def __init__(self,
                 text_cleaner: cln.TextCleaner = None,
                 text_pre_process: pp.TextPreprocessor = None):
        self.text_cleaner = text_cleaner
        self.text_pre_process = text_pre_process


class TimeSeriesHandler:
    def __init__(self,
                 time_frequency=None,
                 n_series_features=3,
                 feature_types=None,
                 id_cols=None,
                 lag_values=None  # Dict of (events back, min/max/avg)
                 ):
        if feature_types is None:
            self.feature_types = ['static', 'freq', 'count', 'lag', 'cyclic']
        else:
            self.feature_types = feature_types
        self.time_frequency = time_frequency
        self.n_series_features = n_series_features
        self.id_cols = id_cols
        if lag_values is None:
            period_back = [1, 2, 3, 5, 7, 14, 21, 360, 1000]
            calc_type = ['mean', 'median', 'min', 'max']
            self.lag_values = period_back, calc_type
        else:
            self.lag_values = lag_values


class FeatureStats:
    def __init__(self, plot=True):
        self.plot = plot


class TextFeatureSpec:
    def __init__(self,
                 split_data: bool = True, split_prop: float = 0.3, split_random_state=42,
                 text_column="text", target_column='target',
                 corex=True, corex_dim=50, tfidf=True, tfidf_dim=100, bow=True, bow_dim=100,
                 ngram_range=(2, 1)):
        self.split_data = split_data
        self.split_prop = split_prop
        self.split_random_state = split_random_state
        self.text_column = text_column
        self.target_column = target_column
        self.corex = corex
        self.corex_dim = corex_dim
        self.tfidf = tfidf
        self.tfidf_dim = tfidf_dim
        self.bow = bow
        self.bow_dim = bow_dim
        self.ngram_range = ngram_range


class FeatureExtraction:
    def __init__(self, text_featrue_extraction: TextFeatureSpec):
        self.text_featrue_extraction = text_featrue_extraction


class PreModel:
    def __init__(self, plot=True):
        self.plot = plot


class ModelRun:
    def __init__(self, best_practice_prob=0.25, n_models=5):
        self.best_practice_prob = best_practice_prob
        self.n_models = n_models


class PostModel:
    def __init__(self, plot=True):
        self.plot = plot


class ModelWeakness:
    def __init__(self, plot=True):
        self.plot = plot



from typing import Dict, Union

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from prettytable import PrettyTable
from textblob import TextBlob

from chester.util import ReportCollector, REPORT_PATH


def report_sentiment_stats(sentiment_df: pd.DataFrame) -> Dict[str, Union[int, float]]:
    rc = ReportCollector(REPORT_PATH)
    # count number of positive, negative and neutral sentiments
    pos_count = len(sentiment_df[sentiment_df['sentiment'] > 0])
    neg_count = len(sentiment_df[sentiment_df['sentiment'] < 0])
    neutral_count = len(sentiment_df[sentiment_df['sentiment'] == 0])
    # calculate percentage of positive, negative and neutral sentiments
    total_count = pos_count + neg_count + neutral_count
    pos_percent = pos_count / total_count * 100
    neg_percent = neg_count / total_count * 100
    neutral_percent = neutral_count / total_count * 100

    # create a table
    table = PrettyTable()
    table.field_names = ["Sentiment", "Count", "Percentage"]
    table.add_row(["Positive", pos_count, f"{pos_percent:.1f}%"])
    table.add_row(["Negative", neg_count, f"{neg_percent:.1f}%"])
    table.add_row(["Neutral", neutral_count, f"{neutral_percent:.1f}%"])

    rc.save_object(obj=table.get_string(title="Sentiment Statistics"), text="Sentiment Statistics")
    return table.get_string(title="Sentiment Statistics")


def analyze_sentiment(df, text_column='text'):
    # Add a new column to the DataFrame with the sentiment of each text
    df['sentiment'] = df[text_column].apply(lambda x: TextBlob(x).sentiment.polarity)

    return df


def plot_sentiment_scores(sentiment_df):
    # Create a histogram of the sentiment scores
    ax = sns.histplot(data=sentiment_df, x='sentiment', kde=False)
    sns.set(rc={'figure.figsize': (12, 12)})
    plt.rcParams.update({'font.size': 14})
    ax.set(xlabel='Sentiment Score', ylabel='Count', title='Sentiment Scores')
    plt.show()
    plt.close()



import pandas as pd


def determine_if_text_or_categorical_column(column):
    # Check if column is a string or categorical data type
    if column.dtype == 'object' or column.dtype.name == 'category':
        # Check if column contains more than 50% unique values
        if len(column.unique()) > 0.5 * len(column):
            return True, False  # Text column
        else:
            return False, True  # Categorical column
    return False, False  # Not a text or categorical column


def test_determine_if_text_column():
    data = {'col1': ['this is text', 'so is this', 'this is not', 1, 2, 3]}
    df = pd.DataFrame(data)
    assert (determine_if_text_or_categorical_column(df['col1']) == (True, False))
    data = {'col1': [1, 2, 3, 4, 5, 6]}
    df = pd.DataFrame(data)
    assert (determine_if_text_or_categorical_column(df['col1']) == (False, False))
    data = {'col1': ['cat', 'dog', 'cat', 'dog', 'cat', 'dog', 'cat', 'dog', 'cat', 'dog']}
    df = pd.DataFrame(data)
    assert (determine_if_text_or_categorical_column(df['col1']) == (False, True))
    import random
    data = {'col1': [random.choice(
        ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
         'w', 'x', 'y', 'z']) for _ in range(1000)]}
    df = pd.DataFrame(data)
    assert (determine_if_text_or_categorical_column(df['col1']) == (False, True))
    data = {'col1': ['cat', 'dog', 'cat', 'dog']}
    df = pd.DataFrame(data)
    assert (determine_if_text_or_categorical_column(df['col1']) == (False, True))



from tamtam.ab_info.ab_class import ABInfo
from tamtam.user_class.user_class import TestInfo


class ABCatboostPlotter:
    def __init__(self, ab_info: ABInfo, test_info: TestInfo, metric):
        self.ab_info = ab_info
        self.test_info = test_info
        self.metric = metric
        self.df = self.ab_info.df
        self.feature_cols = self.test_info.feature_cols

    def preprocess_df(self):
        # df[metrics] = df[metrics].multiply(df[side_col].map({'A': -1, 'B': 1}), axis=0)
        pass

    def run(self):
        pass



import itertools
import re
from datetime import datetime

import pandas as pd
# need an extension for ts module
from dateutil.parser import parse

from chester.zero_break.text_detector import determine_if_text_or_categorical_column


def is_date(string):
    if len(string) < 7:
        return False
    # Define regular expressions to match various date and datetime formats
    date_formats = ["%Y-%m-%d", "%m/%d/%Y", "%d/%m/%Y"]
    datetime_formats = ["%Y-%m-%d %H:%M:%S", "%Y-%m-%d %H:%M", "%m/%d/%Y %H:%M:%S", "%m/%d/%Y %H:%M"]
    regexes = [
        re.compile(r'\d{4}-\d{2}-\d{2}'),
        re.compile(r'\d{2}/\d{2}/\d{4}'),
        re.compile(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}'),
        re.compile(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}'),
        re.compile(r'\d{2}/\d{2}/\d{4} \d{2}:\d{2}:\d{2}'),
        re.compile(r'\d{2}/\d{2}/\d{4} \d{2}:\d{2}')
    ]

    # Check if the string matches any of the known formats or regexes
    try:
        parse(string)
        return True
    except ValueError:
        for fmt in date_formats + datetime_formats:
            try:
                datetime.strptime(string, fmt)
                return True
            except ValueError:
                pass
        for regex in regexes:
            if regex.match(string):
                return True
        return False


class DataInfo:
    def __init__(self, data: pd.DataFrame, target: str = None):
        self.data = data
        self.target = target
        self.is_model = False if self.target is None else True
        self.problem_type_val = None
        self.feature_types_val = None
        self.loss_detector_val = None
        self.metrics_detector_val = None
        self.model_selection_val = None
        self.label_transformation_val = None
        self.rows = self.data.shape[0]
        self.feature_types_val_flatten = None

    def get_all_features(self):
        return list(set(itertools.chain(*self.feature_types_val.values())))

    def calculate(self):
        self.problem_type_val = self.problem_type()
        self.feature_types_val = self.feature_types()
        self.loss_detector_val = self.loss_detector()
        self.metrics_detector_val = self.metrics_detector()
        self.model_selection_val = self.model_selection()
        self.label_transformation_val = self.label_transformation()

    def has_target(self) -> bool:
        return self.target is not None

    def problem_type(self):
        if self.target is None:
            return "No target variable"
        elif pd.api.types.is_numeric_dtype(self.data[self.target]):
            if self.data[[self.target]].drop_duplicates().shape[0] == 2:
                return "Binary regression"
            else:
                return "Regression"
        elif self.data[[self.target]].drop_duplicates().shape[0] == 2:
            return "Binary classification"
        else:
            return "Multiclass classification"

    def _determine_time_cols(self):
        time_cols = []
        for col in self.data.columns:
            if self.data[col].dtype in ['int64', 'float64']:
                pass
            elif col == self.target:
                pass
            elif pd.api.types.is_datetime64_dtype(self.data[col]):
                time_cols.append(col)
            else:
                non_missing_values = self.data[col][self.data[col].notna()].astype(str)
                if non_missing_values.shape[0] == 0:
                    pass
                count = 0
                for value in non_missing_values:
                    if is_date(value):
                        count += 1
                if count / non_missing_values.shape[0] >= 0.9:
                    time_cols.append(col)
        return time_cols

    def _determine_numerical_cols(self):
        numerical_cols = []
        for col in self.data.columns:
            if self.data[col].dtype in ['int64', 'float64']:
                numerical_cols.append(col)
        return numerical_cols

    def _determine_boolean_cols(self):
        boolean_cols = []
        for col in self.data.columns:
            if pd.api.types.is_bool_dtype(self.data[col]):
                boolean_cols.append(col)
        return boolean_cols

    def _determine_id_cols(self):
        id_cols = []
        for col in self.data.columns:
            if col.startswith("ID_") or col.endswith("_id") \
                    or col.endswith("_ID") or col.lower().endswith(" id") \
                    or col.lower() == "id":
                id_cols.append(col)
        return id_cols

    @staticmethod
    def remove_common_elements(l1, l2):
        return [x for x in l1 if x not in l2]

    def feature_types(self):
        # get basic
        numerical_cols = self._determine_numerical_cols()
        boolean_cols = self._determine_boolean_cols()
        time_cols = self._determine_time_cols()
        id_cols = self._determine_id_cols()

        text_cols = []
        categorical_cols = []
        for col in self.data.columns:
            if col not in numerical_cols and col not in boolean_cols:
                is_text, is_categorical = determine_if_text_or_categorical_column(self.data[col])
                if is_text:
                    text_cols.append(col)
                elif is_categorical:
                    categorical_cols.append(col)
        # remove target
        if self.target in numerical_cols:
            numerical_cols.remove(self.target)
        if self.target in boolean_cols:
            boolean_cols.remove(self.target)
        if self.target in text_cols:
            text_cols.remove(self.target)
        if self.target in categorical_cols:
            categorical_cols.remove(self.target)

        # hirerchy
        numerical_cols = list(self.remove_common_elements(numerical_cols, time_cols + id_cols))
        time_cols = list(self.remove_common_elements(time_cols, id_cols))
        text_cols = list(self.remove_common_elements(text_cols, time_cols + id_cols))
        categorical_cols = list(self.remove_common_elements(categorical_cols, time_cols + id_cols))
        return {'numeric': numerical_cols, 'boolean': boolean_cols, 'text': text_cols,
                'categorical': categorical_cols, 'time': time_cols, 'id': id_cols}

    def loss_detector(self):
        problem_type = self.problem_type()
        if problem_type == "Regression":
            return "R squared"
        elif problem_type in ["Binary regression", "Binary classification", "Multiclass classification"]:
            return "Cross entropy"
        else:
            return None

    def metrics_detector(self):
        problem_type = self.problem_type()
        if problem_type == "No target variable":
            return None
        elif problem_type == "Binary regression":
            return ["MSE", "ROC", "MAE", "MAPE"]
        elif problem_type == "Binary classification":
            return ["Accuracy", "Precision", "Recall", "F1"]
        elif problem_type == "Regression":
            return ["R2", "MSE", "MAE", "MAPE"]
        elif problem_type == "Multiclass classification":
            return ["Accuracy", "Precision", "Recall", "F1"]

    def model_selection(self):
        problem_type = self.problem_type()
        if problem_type == "No target variable":
            return None
        elif problem_type == "Binary regression":
            return {"linear", "logistic", "catboost", "baseline-median", "baseline-average"}
        elif problem_type == "Regression":
            return {"linear", "catboost", "baseline-median", "baseline-average"}
        elif problem_type == "Binary classification":
            return {"logistic", "catboost", "baseline-mode"}
        elif problem_type == "Multiclass classification":
            return {"logistic", "catboost", "baseline-mode"}

    def label_transformation(self):
        if self.problem_type() == "No target variable":
            return None
        elif self.problem_type() != "Regression":
            return None
        else:
            return ["Winsorizing", "Log Transformation", "Square Root Transformation", "Reciprocal transformation"]

    def __str__(self):
        report = "Data Information Report\n"
        if self.problem_type_val:
            report += "Problem Type: " + self.problem_type_val + "\n"
        if self.target:
            report += "Target Variable: " + self.target + "\n"
        report += "Feature Types: " + str(self.feature_types_val) + "\n"
        if self.target:
            report += "Loss Function: " + self.loss_detector_val + "\n"
        report += "Evaluation Metrics: " + str(self.metrics_detector_val) + "\n"
        report += "Optional Models: " + str(self.model_selection_val) + "\n"
        # report += "Label Transformation: " + str(self.label_transformation_val)
        return report



import setuptools

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

setuptools.setup(
    name="madcat",
    version="0.63",
    author="Amit Osi",
    author_email="amitosi6666@gmail.com",
    description=open('README.md').read(),
    url="https://github.com/amito-ds/chester",
    packages=setuptools.find_packages(exclude=["tests*"]),
    include_package_data=True,
    python_requires=">=3.8",
    # List of dependencies for this package
    install_requires=install_requires,
    setup_requires=["pytest-runner"]
)



import numpy as np
import pandas as pd
import seaborn as sns
from catboost import CatBoostRegressor
from matplotlib import pyplot as plt
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import OneHotEncoder
from sklearn.tree import DecisionTreeRegressor, export_text

from chester.feature_stats.utils import create_pretty_table
from chester.model_training.data_preparation import CVData
from chester.zero_break.problem_specification import DataInfo


class ModelWeaknesses:
    def __init__(self, cv_data: CVData, data_info: DataInfo, model):
        self.cv_data = cv_data
        self.data_info = data_info
        self.model = model
        self.X_train = self.cv_data.train_data.drop(columns=[self.cv_data.target_column])
        self.y_train = self.cv_data.train_data[self.cv_data.target_column]
        self.X_test = self.cv_data.test_data.drop(columns=[self.cv_data.target_column])
        self.y_test = self.cv_data.test_data[self.cv_data.target_column]
        # retrain the model
        self.model.retrain(self.X_train, self.y_train)
        # predicting
        self.predict_test = self.model.predict(self.X_test)
        self.predict_test = self.handle_predictions()
        self.error = self.calculate_error()
        from chester.util import ReportCollector, REPORT_PATH
        self.rc = ReportCollector(REPORT_PATH)

    def handle_predictions(self):
        y = self.predict_test
        if isinstance(y, pd.Series):
            return y
        elif y.ndim == 2 and y.shape[1] == 1:
            return pd.Series(y.flatten()).reset_index(drop=True)
        elif y.ndim == 2 and y.shape[1] == 2:
            return pd.Series(np.argmax(y, axis=1))
        return y

    def calculate_error(self):
        if 'class' in self.data_info.problem_type_val.lower():
            y_test = self.y_test.reset_index(drop=True)
            try:
                predict_test = self.predict_test.reset_index(drop=True)
            except:
                predict_test = self.predict_test
            error_per_row = 1 - (y_test == predict_test)
            return pd.Series(error_per_row, name='error')
        elif 'regression' in self.data_info.problem_type_val.lower():
            y_test = self.y_test.reset_index(drop=True)
            try:
                predict_test = self.predict_test.reset_index(drop=True)
            except:
                predict_test = self.predict_test
            smape = 2 * np.abs(y_test - predict_test) / (np.abs(y_test) + np.abs(predict_test))
            smape = smape.replace([np.inf, -np.inf], np.nan)
            smape = smape.dropna()
            return pd.Series(smape, name='error')

    def plot_decision_tree_error_regressor(self, min_samples_leaf=0.2, max_depth=2):
        from sklearn import tree
        from matplotlib import pyplot as plt

        # Prepare the numerical features for processing
        numerical_transformer = SimpleImputer(strategy='median')
        # Prepare the categorical features for processing
        categorical_transformer = OneHotEncoder(handle_unknown='ignore')
        # Use the ColumnTransformer to apply the transformations to the corresponding features
        transformer = ColumnTransformer(
            transformers=[
                ('c', categorical_transformer, self.data_info.feature_types_val["categorical"]),
                ('n', numerical_transformer, self.data_info.feature_types_val["numeric"])
            ])
        # Apply the transformations to the X_test data
        X_test_transformed = transformer.fit_transform(self.X_test)

        # Train the DecisionTreeRegressor model
        model = DecisionTreeRegressor(min_samples_leaf=min_samples_leaf, max_depth=max_depth)
        model.fit(X_test_transformed, self.error)

        # Get the feature names of the numeric features
        numeric_feature_names = self.data_info.feature_types_val["numeric"]
        # Get the feature names of the categorical features
        categorical_transformer.fit(self.X_test[self.data_info.feature_types_val["categorical"]])
        categories = categorical_transformer.categories_
        categorical_feature_names = []
        for i, col in enumerate(self.data_info.feature_types_val["categorical"]):
            for j in range(len(categories[i])):
                categorical_feature_names.append(f"{col}_{categories[i][j]}")
        # Concatenate the feature names of the categorical and numeric features
        feature_names = numeric_feature_names + categorical_feature_names

        plt.figure(figsize=(13, 13))
        tree.plot_tree(model,
                       feature_names=feature_names,
                       class_names=['error'],
                       rounded=True,
                       filled=True,
                       )
        # Get the tree structure as text
        self.rc.save_object(obj=export_text(model, feature_names=feature_names),
                            text="Print of tree to look for potential segments with high error (use with caution)")
        plt.suptitle("Decision Tree Trained on Model Error")

    def plot_catboost_error_regressor(self, iterations=100, depth=3, learning_rate=0.1):
        model = CatBoostRegressor(iterations=iterations, depth=depth, learning_rate=learning_rate)
        model.fit(self.X_test, self.error, verbose=False)
        plt.figure(figsize=(15, 15))
        feature_importances = np.round(model.feature_importances_, decimals=0)
        feature_imp = pd.DataFrame({'Feature': self.X_test.columns, 'Importance': feature_importances})
        feature_imp = feature_imp.sort_values(by='Importance', ascending=False)
        self.rc.save_object(create_pretty_table(feature_imp[0:50]), text="Most important features by catboost:")
        feature_imp = feature_imp[0:30]
        sns.barplot(x=feature_imp['Importance'], y=feature_imp['Feature'])
        plt.title('CatBoost Feature Importance to Detect Segments with High Error')
        plt.show()
        plt.close()

    def run(self):
        print("Training model to predict the error")
        if np.unique(self.error).size == 1:
            print("🎉 No weaknesses found! All errors on the test set are 0.")
            return None
        self.plot_catboost_error_regressor()
        self.plot_decision_tree_error_regressor()






import numpy as np
import os

import numpy as np
import pandas as pd
# exec(open("/Users/amitosi/PycharmProjects/chester/diamond/manual_run.py").read())
from PIL import Image

from diamond.run import run
from diamond.user_classes import ImagesAugmentationInfo, ImageModel, ImageModels

# # define directory path
data_dir = '/Users/amitosi/PycharmProjects/chester/chester/data/pizza_not_pizza'
# subdir: not pizza, .jpg
# subdir: pizza, .jpg

# # define image size
img_size = (64, 64)

# define empty arrays to store images and labels
images = []
labels = []

# loop over the directories containing the images

# loop over the subdirectories
for subdir in os.listdir(data_dir):
    # get the full path of the subdirectory
    subdir_path = os.path.join(data_dir, subdir)

    # loop over the files in the subdirectory
    for filename in os.listdir(subdir_path):
        # get the full path of the file
        file_path = os.path.join(subdir_path, filename)

        # load the image using PIL
        img = Image.open(file_path)

        # resize the image to img_size
        img = img.resize(img_size)

        # convert the image to a numpy array
        img = np.array(img)

        # append the image to the images array
        images.append(img)

        # append the label to the labels array
        if subdir == 'pizza':
            labels.append(1)
        else:
            labels.append(0)

# convert the images and labels arrays to numpy arrays
sample_indices = np.random.choice(len(images), size=50, replace=False)
images = np.array(images)[sample_indices]
labels = np.array(labels)[sample_indices]


image_shape = (3, 64, 64)


image_model_list = [
    ImageModel(network_name="EfficientNetB0",
               batch_size=64 * 64 * 16,
               num_epochs=1,
               optimizer_params={'lr': 0.005},
               dropout=0.7)]
image_models = ImageModels(image_model_list=image_model_list)

diamond_collector = run(images=images,
                        image_shape=image_shape,
                        labels=labels,
                        get_image_description=True,
                        is_augment_data=False,
                        image_augmentation_info=ImagesAugmentationInfo(aug_prop=0.1),
                        is_train_model=False, image_models=image_models,
                        is_post_model_analysis=False,
                        plot=True)



from tamtam.utils.utils import true_if_exists


class ABData:
    def __init__(self, df):
        self.df = df


class TestInfo:
    def __init__(self, side_col, metrics: list,
                 id_cols=None, feature_cols=None, weight_col=None, date_col=None, is_higher_better=True,
                 trimming_list=None):
        self.side_col = side_col
        self.metrics = metrics
        self.id_cols = id_cols
        self.feature_cols = feature_cols
        self.weight_col = weight_col
        self.date_col = date_col
        self.is_higher_better = is_higher_better
        self.trimming_list = trimming_list
        if not self.trimming_list:
            self.trimming_list = [0, 0.001, 0.005, 0.01, 0.02, 0.05, 0.1]

    def get_side_col(self):
        if not self.side_col:
            return []
        return [self.side_col]

    def get_metrics(self):
        if not self.metrics:
            return []
        return self.metrics

    def get_id_cols(self):
        if not self.id_cols:
            return []
        return self.id_cols

    def is_id_cols(self):
        return true_if_exists(self.id_cols)

    def get_feature_cols(self):
        if not self.feature_cols:
            return []
        return self.feature_cols

    def get_weight_col(self):
        return self.weight_col

    def is_weight_col(self):
        return true_if_exists(self.weight_col)

    def get_date_col(self):
        if not self.date_col:
            return []
        return self.date_col

    def get_is_higher_better(self):
        return self.is_higher_better

    def get_all_cols(self):
        all_cols = self.get_side_col() + self.get_metrics() + self.get_id_cols() + self.get_feature_cols()
        if self.get_weight_col():
            all_cols.append(self.get_weight_col())
        if self.get_date_col():
            all_cols.append(self.get_date_col())
        return all_cols
